from typing import Optional
from unittest import TestCase

from flask import Blueprint, Flask

from abstra_internals.jwt_auth import USER_AUTH_HEADER_KEY, UserClaims
from abstra_internals.repositories.project.project import (
    AccessSettings,
    FormStage,
    NotificationTrigger,
    ProjectRepository,
    WorkflowTransition,
)
from abstra_internals.repositories.users import TestUsersRepository
from abstra_internals.server.guards.role_guard import (
    Guard,
    IdArgSelector,
    PathArgSelector,
    StageIdSelector,
)
from tests.fixtures import clear_dir, init_dir


class TestRequirementsApi(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_app_protected_by(self):
        ##Autorization: Bearer <test_email>
        def test_header(email: str):
            return {USER_AUTH_HEADER_KEY: f"Bearer {email}"}

        def test_auth_decoder(email: str):
            email = email.split(" ")[1]
            return UserClaims(claims={"email": email})

        JOB_EXECUTOR = "job_executor"
        WORKFLOW_VIEWER = "workflow_viewer"
        ADMIN = "admin"

        TESTER_EMAIL = "tester@abstra.app"
        ADMIN_EMAIL = "admin@abstra.app"
        USER_EMAIL = "user@abstra.app"

        NOT_A_USER_EMAIL = "external@evil.corp"

        repo = TestUsersRepository(
            {
                TESTER_EMAIL: [WORKFLOW_VIEWER],
                ADMIN_EMAIL: [ADMIN, JOB_EXECUTOR, WORKFLOW_VIEWER],
                USER_EMAIL: [],
            }
        )

        project = ProjectRepository.load()

        public_form = FormStage(
            id="public_form_id",
            path="public_form",
            title="public_form",
            file="public_form.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
            access_control=AccessSettings(is_public=True, required_roles=[]),
        )

        public_form = FormStage(
            id="public_form_id",
            path="public_form",
            title="public_form",
            file="public_form.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
            access_control=AccessSettings(is_public=True, required_roles=[]),
        )

        protected_form = FormStage(
            id="protected_form_id",
            path="protected_form",
            title="protected_form",
            file="protected_form.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
            access_control=AccessSettings(is_public=False, required_roles=[]),
        )

        private_form = FormStage(
            id="private_form_id",
            path="private_form",
            title="private_form",
            file="private_form.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
            access_control=AccessSettings(is_public=False, required_roles=[ADMIN]),
        )

        project.forms.append(public_form)
        project.forms.append(protected_form)
        project.forms.append(private_form)

        ProjectRepository.save(project)

        guard = Guard(repo, enabled=True, auth_decoder=test_auth_decoder)

        app = Flask(__name__)
        bp = Blueprint("dummy", __name__)

        @bp.route("/protected_form_id")
        @guard.by(StageIdSelector("protected_form_id"))
        def protected_form_route():
            return {"message": "OK"}

        @bp.route("/_pages/<string:path>")
        @guard.by(PathArgSelector("path"))
        def path_stage(path: str):
            return {"message": "OK", "path": path}

        @bp.route("/stage/<path:id>/start")
        @guard.by(IdArgSelector("id"))
        def id_stage(id: str):
            return {"message": "OK", "id": id}

        app.register_blueprint(bp, url_prefix="/")

        ## Logged in users
        test_matrix = [
            {
                "email": TESTER_EMAIL,
                "test_routes": {
                    "/protected_form_id": "200 OK",
                    "/_pages/public_form": "200 OK",
                    "/_pages/protected_form": "200 OK",
                    "/_pages/private_form": "403 FORBIDDEN",
                    "/stage/public_form_id/start": "200 OK",
                    "/stage/protected_form_id/start": "200 OK",
                    "/stage/private_form_id/start": "403 FORBIDDEN",
                },
            },
            {
                "email": ADMIN_EMAIL,
                "test_routes": {
                    "/protected_form_id": "200 OK",
                    "/_pages/public_form": "200 OK",
                    "/_pages/protected_form": "200 OK",
                    "/_pages/private_form": "200 OK",
                    "/stage/public_form_id/start": "200 OK",
                    "/stage/protected_form_id/start": "200 OK",
                    "/stage/private_form_id/start": "200 OK",
                },
            },
            {
                "email": USER_EMAIL,
                "test_routes": {
                    "/protected_form_id": "200 OK",
                    "/_pages/public_form": "200 OK",
                    "/_pages/protected_form": "200 OK",
                    "/_pages/private_form": "403 FORBIDDEN",
                    "/stage/public_form_id/start": "200 OK",
                    "/stage/protected_form_id/start": "200 OK",
                    "/stage/private_form_id/start": "403 FORBIDDEN",
                },
            },
            {
                "email": NOT_A_USER_EMAIL,
                "test_routes": {
                    "/protected_form_id": "403 FORBIDDEN",
                    "/_pages/public_form": "200 OK",
                    "/_pages/protected_form": "403 FORBIDDEN",
                    "/_pages/private_form": "403 FORBIDDEN",
                    "/stage/public_form_id/start": "200 OK",
                    "/stage/protected_form_id/start": "403 FORBIDDEN",
                    "/stage/private_form_id/start": "403 FORBIDDEN",
                },
            },
        ]

        client = app.test_client()
        for case in test_matrix:
            for route, status in case["test_routes"].items():
                response = client.get(route, headers=test_header(case["email"]))
                self.assertEqual(
                    response.status, status, f"Failed for {route} with {case['email']}"
                )

        ## Logged off user
        test_matrix = {
            "/protected_form_id": "401 UNAUTHORIZED",
            "/_pages/public_form": "200 OK",
            "/_pages/protected_form": "401 UNAUTHORIZED",
            "/_pages/private_form": "401 UNAUTHORIZED",
            "/stage/public_form_id/start": "200 OK",
            "/stage/protected_form_id/start": "401 UNAUTHORIZED",
            "/stage/private_form_id/start": "401 UNAUTHORIZED",
        }

        for route, status in test_matrix.items():
            response = client.get(route, headers={})
            self.assertEqual(response.status, status, f"Failed for {route}")

    def test_sidebar(self):
        ##Autorization: Bearer <test_email>
        def test_auth(email: Optional[str]):
            if email is None:
                return None
            return f"Bearer {email}"

        def test_auth_decoder(email: str):
            email = email.split(" ")[1]
            return UserClaims(claims={"email": email})

        JOB_EXECUTOR = "job_executor"
        WORKFLOW_VIEWER = "workflow_viewer"
        ADMIN = "admin"

        TESTER_EMAIL = "tester@abstra.app"
        ADMIN_EMAIL = "admin@abstra.app"
        USER_EMAIL = "user@abstra.app"

        NOT_A_USER_EMAIL = "external@evil.corp"

        repo = TestUsersRepository(
            {
                TESTER_EMAIL: [WORKFLOW_VIEWER],
                ADMIN_EMAIL: [ADMIN, JOB_EXECUTOR, WORKFLOW_VIEWER],
                USER_EMAIL: [],
            }
        )

        project = ProjectRepository.load()

        public_form = FormStage(
            id="public_form_id",
            path="public_form",
            title="public_form",
            file="public_form.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
            access_control=AccessSettings(is_public=True, required_roles=[]),
        )

        public_pependent_form = FormStage(
            id="public_pependent_form",
            path="public_pependent_form",
            title="public_pependent_form",
            file="public_pependent_form.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
            access_control=AccessSettings(is_public=True, required_roles=[]),
        )

        public_form.workflow_transitions.append(
            WorkflowTransition(
                id="1",
                target_type="target_type",
                target_id="public_pependent_form",
                type="form:finished",
            )
        )

        protected_form = FormStage(
            id="protected_form_id",
            path="protected_form",
            title="protected_form",
            file="protected_form.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
            access_control=AccessSettings(is_public=False, required_roles=[]),
        )

        private_form = FormStage(
            id="private_form_id",
            path="private_form",
            title="private_form",
            file="private_form.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
            access_control=AccessSettings(is_public=False, required_roles=[ADMIN]),
        )

        project.forms.append(public_form)
        project.forms.append(public_pependent_form)
        project.forms.append(protected_form)
        project.forms.append(private_form)

        ProjectRepository.save(project)

        guard = Guard(repo, enabled=True, auth_decoder=test_auth_decoder)

        ## Logged in users
        test_matrix = [
            {
                "email": TESTER_EMAIL,
                "accessible_stages_ids": [
                    "home",
                    "public_form_id",
                    "protected_form_id",
                ],
            },
            {
                "email": ADMIN_EMAIL,
                "accessible_stages_ids": [
                    "home",
                    "public_form_id",
                    "protected_form_id",
                    "private_form_id",
                ],
            },
            {
                "email": USER_EMAIL,
                "accessible_stages_ids": [
                    "home",
                    "public_form_id",
                    "protected_form_id",
                ],
            },
            {
                "email": NOT_A_USER_EMAIL,
                "accessible_stages_ids": [],
            },
            {
                "email": None,  ##Logged off user
                "accessible_stages_ids": [],
            },
        ]

        for case in test_matrix:
            auth = test_auth(case["email"])
            workspace = guard.filtered_workspace(auth)
            response = [stage.id for stage in workspace.sidebar.items]
            expected = case["accessible_stages_ids"]

            self.assertEqual(response, expected, f"Failed for {case['email']}")



from typing import List

from abstra_internals.controllers.sdk_context import SDKContextStore
from abstra_internals.controllers.sdk_tasks import Task
from abstra_internals.repositories.tasks import TaskPayload
from abstra_internals.utils.json import to_json_serializable


def get_tasks(limit=100) -> List[Task]:
    context = SDKContextStore.get_by_thread()
    return context.task_sdk.get_stage_pending_tasks(limit)


def send_task(type: str, payload: TaskPayload, show_warning=True) -> None:
    context = SDKContextStore.get_by_thread()
    if not isinstance(payload, dict):
        raise Exception("Task payload must be a dictionary")
    serializable_payload = {k: to_json_serializable(v) for k, v in payload.items()}
    context.task_sdk.send_task(type, serializable_payload, show_warning=show_warning)


def get_trigger_task() -> Task:
    context = SDKContextStore.get_by_thread()
    return context.task_sdk.get_trigger_task()



from typing import Dict, List, Union

import flask
from pydantic import Field

from abstra_internals.utils.dict import filter_non_string_values
from abstra_internals.utils.serializable import Serializable


class Request(Serializable):
    query_params: Dict[str, str]
    headers: Dict[str, str]
    method: str
    body: str


class Response(Serializable):
    headers: Dict[str, str]
    status: int
    body: str


class HookContext(Serializable):
    request: Request
    response: Response
    sent_tasks: List[str] = Field(default_factory=list)
    legacy_thread_data: dict = Field(default_factory=dict)


class FormContext(Serializable):
    request: Request
    sent_tasks: List[str] = Field(default_factory=list)
    legacy_thread_data: dict = Field(default_factory=dict)


class ScriptContext(Serializable):
    task_id: str
    sent_tasks: List[str] = Field(default_factory=list)
    legacy_thread_data: dict = Field(default_factory=dict)


class JobContext(Serializable):
    sent_tasks: List[str] = Field(default_factory=list)
    legacy_thread_data: dict = Field(default_factory=dict)


ClientContext = Union[HookContext, FormContext, ScriptContext, JobContext]


def extract_flask_request(request: flask.Request) -> Request:
    return Request(
        headers=filter_non_string_values(request.headers),
        body=request.get_data(as_text=True),
        query_params={**request.args},
        method=request.method,
    )



import flask

from abstra_internals.controllers.ai import AiController
from abstra_internals.controllers.main import MainController
from abstra_internals.usage import editor_manual_usage, editor_usage


def get_editor_bp(main_controller: MainController):
    bp = flask.Blueprint("editor_ai", __name__)
    controller = AiController(main_controller)

    @bp.post("/message")
    @editor_usage
    def _get_next_message():
        body = flask.request.json
        if not body:
            flask.abort(400)

        streamer = controller.send_ai_message(
            body["messages"], body["runtime"], body["threadId"]
        )

        if streamer is None:
            flask.abort(403)

        return flask.Response(streamer, mimetype="text/event-stream")

    @bp.post("/thread")
    @editor_usage
    def _create_thread():
        thread = controller.create_thread()
        if not thread:
            flask.abort(403)
        return thread

    @bp.post("/cancel-all")
    @editor_usage
    def _cancel_all():
        body = flask.request.json
        if not body:
            flask.abort(400)
        thread_id = body.get("threadId")
        if not thread_id:
            flask.abort(400)
        controller.cancel_all(thread_id)
        return {"success": True}

    @bp.post("/generate")
    @editor_usage
    def _generate():
        body = flask.request.json
        if not body:
            flask.abort(400)

        prompt = body.get("prompt")
        if not prompt:
            flask.abort(400)

        controller.generate_project(prompt)
        return {"success": True}

    @bp.post("/vote")
    def _vote():
        body = flask.request.json
        if not body:
            flask.abort(400)
        vote = body.get("vote")
        question = body.get("question")
        answer = body.get("answer")
        context = body.get("context")

        editor_manual_usage(
            event="ai_vote",
            payload=dict(vote=vote, question=question, answer=answer, context=context),
        )

        return {"success": True}

    return bp



from typing import Dict, List, Optional

from abstra_internals.contract.forms import ValidationResult
from abstra_internals.controllers.execution_client_form import FormClient
from abstra_internals.interface.sdk import user_exceptions
from abstra_internals.jwt_auth import UserClaims
from abstra_internals.repositories.users import UsersRepository


class FormSDKController:
    def __init__(self, client: FormClient, users_repository: UsersRepository) -> None:
        self.client = client
        self.users_repository = users_repository

    def get_user(self, force_refresh: bool) -> UserClaims:
        data = self.client.request_auth(force_refresh)
        claims = UserClaims.from_jwt(data["jwt"])

        if not claims:
            self.client.handle_invalid_jwt()
            raise user_exceptions.GetUserFailed()

        if user := self.users_repository.get_user(claims.email):
            claims.add_roles(user.roles)

        self.client.handle_valid_jwt()
        return claims

    def execute_js(self, code: str, context: dict):
        return self.client.request_execute_js(code, context)

    def redirect(self, url: str, query_params: Optional[Dict[str, str]] = None) -> None:
        self.client.request_redirect(url, query_params)

    def get_query_params(self) -> Dict[str, str]:
        return self.client.get_query_params()

    def request_mount_page(
        self,
        widgets: list,
        actions: list,
        end_program: bool,
        reactive_polling_interval: int,
        steps_info: Optional[List],
    ):
        self.client.request_mount_page(
            widgets,
            actions,
            end_program,
            reactive_polling_interval,
            steps_info,
        )

    def request_page_update(
        self, widgets: list, validation: ValidationResult, event_seq: int
    ):
        self.client.request_page_update(widgets, validation, event_seq)

    def next_message(self) -> Dict:
        return self.client.next_message()



from typing import List, Tuple

from abstra_internals.email_templates import task_waiting_template
from abstra_internals.entities.execution import Execution, PreExecution
from abstra_internals.entities.execution_context import ScriptContext
from abstra_internals.interface.sdk import user_exceptions
from abstra_internals.interface.sdk.user_exceptions import (
    TaskNotWaiting,
    TaskWriteAttempt,
)
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.factory import Repositories
from abstra_internals.repositories.project.project import (
    FormStage,
    Project,
    ProjectRepository,
    ScriptStage,
    Stage,
)
from abstra_internals.repositories.tasks import TaskDTO, TaskLockFailed, TaskPayload


class Task:
    def __init__(self, task_controller: "TasksSDKController", dto: TaskDTO) -> None:
        self._dto = dto
        self._controller = task_controller

    def __str__(self) -> str:
        return str(self._dto)

    def __repr__(self) -> str:
        return self._dto.__repr__()

    def __getitem__(self, key: str) -> object:
        return self._dto.payload[key]

    def __getattr__(self, key: str) -> object:
        return self._dto.payload[key]

    def __setitem__(self, key: str, value: str) -> None:
        raise TaskWriteAttempt()

    @property
    def id(self) -> str:
        return self._dto.id

    def get_payload(self) -> TaskPayload:
        return self._dto.payload

    def get_dto(self) -> TaskDTO:
        return self._dto

    def lock(self) -> None:
        try:
            self._controller.lock_task(self._dto.id)
        except TaskLockFailed:
            raise TaskNotWaiting()

    def complete(self) -> None:
        try:
            self._controller.complete_task(self._dto.id)
        except TaskLockFailed:
            raise TaskNotWaiting()

    def __enter__(self):
        self.lock()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.complete()
        else:
            self._controller.set_task_to_pending(self._dto.id)
            return False


class TasksSDKController:
    def __init__(self, repositories: Repositories, execution: Execution) -> None:
        self.execution = execution
        self.repositories = repositories
        self.project = ProjectRepository().load()

    def _get_task_targets(
        self,
        current_stage: Stage,
        type: str,
        task_payload: TaskPayload,
        project: Project,
        show_warning: bool = True,
    ) -> List[Tuple[Stage, dict]]:
        targets = []

        possible_transitions = list(
            filter(
                lambda t: True if not t.task_type else type == t.task_type,
                current_stage.workflow_transitions,
            )
        )

        for transition in possible_transitions:
            target_stage = project.get_stage(transition.target_id)
            if not target_stage:
                continue

            targets.append((target_stage, task_payload))

        if len(targets) == 0 and show_warning:
            AbstraLogger.warning(
                f"Task with payload {task_payload} was deleted, no transitions were satisfied with the type: {type}"
            )

        return targets

    def send_task(self, type: str, payload: TaskPayload, show_warning=True) -> None:
        current_stage = self.project.get_stage(self.execution.stage_id)

        if not current_stage:
            raise Exception(f"Stage {self.execution.stage_id} not found")

        target_stages = self._get_task_targets(
            current_stage, type, payload, self.project, show_warning=show_warning
        )

        for stage, task_payload in target_stages:
            task = self.repositories.tasks.send_task(
                type, task_payload, stage.id, current_stage.id, self.execution.id
            )
            self._send_waiting_thread_notification(task)
            self.execution.context.sent_tasks.append(task.id)

            if isinstance(stage, ScriptStage):
                self.repositories.producer.submit(
                    PreExecution(
                        context=ScriptContext(task_id=task.id),
                        stage_id=stage.id,
                    )
                )

    def get_stage_pending_tasks(self, limit: int) -> List[Task]:
        stage = self.project.get_stage(self.execution.stage_id)
        if not stage:
            raise Exception(f"Stage {self.execution.stage_id} not found")

        dtos = self.repositories.tasks.get_pending_tasks(self.execution.stage_id, limit)
        return [Task(self, dto) for dto in dtos]

    def lock_task(self, task_id: str) -> None:
        self.repositories.tasks.lock_task(
            task_id, self.execution.id, self.execution.stage_id
        )

    def complete_task(self, task_id: str) -> None:
        self.repositories.tasks.complete_task(
            task_id, self.execution.id, self.execution.stage_id
        )

    def set_task_to_pending(self, task_id: str) -> None:
        self.repositories.tasks.set_task_to_pending(task_id)

    def get_all_tasks(self) -> List[Task]:
        dtos = self.repositories.tasks.get_all_tasks()
        return [Task(self, dto) for dto in dtos]

    def get_trigger_task(self) -> Task:
        if not isinstance(self.execution.context, ScriptContext):
            raise user_exceptions.IncompatibleScriptSDK()

        dto = self.repositories.tasks.get_by_id(self.execution.context.task_id)
        return Task(self, dto)

    def _send_waiting_thread_notification(self, task: TaskDTO):
        stage = self.project.get_stage(task.target_stage_id)
        if not stage:
            raise Exception(f"Stage {task.target_stage_id} not found")

        if not (isinstance(stage, FormStage) and stage.notification_trigger.enabled):
            return

        recipient_emails = stage.notification_trigger.get_recipients(task.payload)
        if not recipient_emails:
            return

        self.repositories.email.send(
            task_waiting_template.generate_email(
                recipient_emails=recipient_emails, form=stage
            )
        )



import importlib
import importlib.util as imp_util
import sys
from pathlib import Path

from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.project.project import ProjectRepository
from abstra_internals.utils.file import path2module


def import_as_new(filepath: str):
    module_spec = imp_util.spec_from_file_location(filepath, filepath)
    if module_spec is None or module_spec.loader is None:
        raise Exception(f"Could not load module {filepath}")

    module = imp_util.module_from_spec(module_spec)
    if module is None:
        raise Exception(f"Could not load module {filepath}")

    if module_spec.cached is not None:
        Path(module_spec.cached).unlink(missing_ok=True)

    module_spec.loader.exec_module(module)

    return module


def reload_project_local_modules():
    try:
        project = ProjectRepository.load()

        for file in project.get_local_dependencies():
            if not file.exists():
                continue

            module_name = path2module(file)
            module = sys.modules.get(module_name)

            try:
                if module is None:
                    importlib.import_module(module_name)
                else:
                    if (
                        module.__spec__ is not None
                        and module.__spec__.cached is not None
                    ):
                        Path(module.__spec__.cached).unlink(missing_ok=True)
                    importlib.reload(module)
            except Exception as e:
                AbstraLogger.error(
                    f"Could not reload module from {file} with the following error: {e}"
                )
                continue
    except Exception:
        pass




from typing import Any, Dict, List, TYPE_CHECKING
from abstra_internals.widgets.types import AbstraOption
from abstra_internals.interface.sdk.forms.page import Page

if TYPE_CHECKING:
    from pandas import DataFrame

def get_single_value(answer: Dict):
    return list(answer.values())[0]


def read_appointment(label: str, **kwargs):
  
    """A calendar slot selection that allows users to select a time slot for an appointment.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		slots (List[Union[AppointmentSlot, dict, Tuple[datetime, datetime]]]): The available slots for the user to choose from
		initial_value (Union[AppointmentSlot, dict, Tuple[datetime, datetime]]): The initial value to display to the user. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A dict containing the selected appointment slot
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_appointment(label, **kwargs).run(button_text))


def read_camera(label: str, **kwargs):
  
    """Captures and uploads a photo using the user's camera.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		multiple (bool): Whether the user will be allowed to upload multiple files. Defaults to False.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A dict containing the picture taken by the user ({"file": file, "url": str, "content": bytes}) or a list of pictures in case of multiple flag set as True
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_camera(label, **kwargs).run(button_text))


def read_cards(label: str, options: list, **kwargs):
  
    """Presents options as cards for the user. Cards may be selectable.

        Position Args: 
		label (str): The text related to this input
		options (list): List of dicts representing the cards, each dict can have the following keys: title, subtitle, image, description, topLeftExtra, topRightExtra
		
        Keyword Args: 
		multiple (bool): Whether the user can select multiple options. Defaults to False.
		initial_value (list): The initial value to display to the user. Defaults to None.
		searchable (bool): Whether to show a search bar. Defaults to False.
		layout (str): Whether the cards layout should be 'list' or 'grid'. Defaults to 'list'. The 'grid' only applies to desktop resolutions.
		columns (int): When layout is 'grid', how many columns to display
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The options/option selected by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_cards(label, options, **kwargs).run(button_text))


def read_checkbox(label: str, **kwargs):
  
    """Enables a selection through a single checkbox interaction.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_checkbox(label, **kwargs).run(button_text))


def read_checklist(label: str, options: List[AbstraOption], **kwargs):
  
    """Displays a checklist allowing multiple item selections.

        Position Args: 
		label (str): The label to display to the user
		options (List[AbstraOption]): The options to display to the user, eg. ['Option 1', 'Option 2'] or [{'label': 'Option 1', 'value': '1'}, {'label': 'Option 2', 'value': '2'}]
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to None.
		min (int): The minimum number of items that must be selected. Defaults to 0.
		max (int): The maximum number of items that can be selected. Defaults to the number of options.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The selected values
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_checklist(label, options, **kwargs).run(button_text))


def read_cnpj(label: str, **kwargs):
  
    """CNPJ number input with validation and masking.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to None.
		invalid_message (str): The message to display when the user enters an invalid value. Defaults to None.
		placeholder (str): The placeholder text to display to the user. Defaults to "00.000.000/0001-00".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_cnpj(label, **kwargs).run(button_text))


def read_code(label: str, **kwargs):
  
    """Accepts code input with syntax highlighting

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		language (str): The programming language. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_code(label, **kwargs).run(button_text))


def read_cpf(label: str, **kwargs):
  
    """CPF number input with validation and masking.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		invalid_message (str): The message to display when the user enters an invalid value. Defaults to None.
		placeholder (str): The placeholder text to display to the user. Defaults to "000.000.000-00".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_cpf(label, **kwargs).run(button_text))


def read_currency(label: str, **kwargs):
  
    """Enables monetary value input with currency formatting.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to 0.
		placeholder (str): The placeholder text to display to the user. Defaults to "".
		min (float): The minimum value allowed, eg. "0". Defaults to None.
		max (float): The maximum value allowed, eg. "100". Defaults to None.
		currency (str): The currency to display to the user, eg. "USD", "BRL, "EUR", "GBP". Defaults to "USD".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_currency(label, **kwargs).run(button_text))


def read_custom(html_body: str, **kwargs):
  
    """Allows for a fully customizable widget with custom HTML, CSS, and JS.

        Position Args: 
		html_body (str): The HTML body content
		
        Keyword Args: 
		initial_value (Any): The initial value to be stored in custom widget state.
		label (str): The label to display to the user
		html_head (str): The HTML head content
		css (str): The widget's CSS
		js (str): The widget's JavaScript
		height (int): The widget's height
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The custom response
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_custom(html_body, **kwargs).run(button_text))


def read_date(label: str, **kwargs):
  
    """Provides a date picker with calendar interface.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (Union[datetime.date, time.struct_time, str]): The initial value to display to the user. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_date(label, **kwargs).run(button_text))


def read_dropdown(label: str, options: List[AbstraOption], **kwargs):
  
    """Provides a dropdown menu for single or multiple selections.

        Position Args: 
		label (str): The label to display to the user
		options (List[AbstraOption]): The options to display to the user, eg. ['Option 1', 'Option 2'] or [{'label': 'Option 1', 'value': '1'}, {'label': 'Option 2', 'value': '2'}]
		
        Keyword Args: 
		multiple (bool): Whether the user can select multiple options. Defaults to False.
		initial_value (str or list): The initial value to display to the user. Defaults to [].
		placeholder (str): The placeholder text to display to the user. Defaults to "Choose an option".
		min (number): The minimal amount of options that should be selected. Defaults to None.
		max (number): The maximum amount of options that should be selected. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value selected by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_dropdown(label, options, **kwargs).run(button_text))


def read_email(label: str, **kwargs):
  
    """ Collects and validates an email address.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		placeholder (str): The placeholder text to display to the user. Defaults to "".
		invalid_email_message (str): Invalid e-mail message
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_email(label, **kwargs).run(button_text))


def read_file(label: str, **kwargs):
  
    """Enables file upload via file explorer.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		multiple (bool): Whether the user will be allowed to upload multiple files. Defaults to False.
		accepted_formats (list): The specific file types that the input should accept. Defaults to [], accepting all file formats.
		min (number): The minimal amount of files that should be submitted. Only applied if multiple is True. Defaults to None.
		max (number): The maximum amount of files that should be submitted. Only applied if multiple is True. Defaults to None.
		max_file_size (float): Maximum size allowed to be transfered in total in MB.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A object containing the file uploaded by the user: FileResponse(path: Path, file: BufferedReader). If the multiple flag is set as True, it might contain a list of FileResponses.
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_file(label, **kwargs).run(button_text))


def read_image(label: str, **kwargs):
  
    """Allows for image file uploads with preview and validation.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		multiple (bool): Whether the user will be allowed to upload multiple files. Defaults to False.
		max_file_size (float): Maximum size allowed to be transfered in total in MB.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A dict containing the image file uploaded by the user: FileResponse(path: Path, file: BufferedReader). If the multiple flag is set as True, it might contain a list of FileResponses.
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_image(label, **kwargs).run(button_text))


def read_list(item_schema: Any, **kwargs):
  
    """Collects a dynamic list of values based on a specified schema.

        Position Args: 
		item_schema (Any): The schema for the items of the list
		
        Keyword Args: 
		initial_value (array): ''
		min (float): Min value accepted by the input. Defaults to None.
		max (float): Max value accepted by the input. Defaults to None.
		add_button_text (str): ''
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The values entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_list(item_schema, **kwargs).run(button_text))


def read_multiple_choice(label: str, options: List[AbstraOption], **kwargs):
  
    """Offers multiple choice selections with single or multi-select options.

        Position Args: 
		label (str): The label to display to the user
		options (List[AbstraOption]): The options to display to the user, eg. ['Option 1', 'Option 2'] or [{'label': 'Option 1', 'value': '1'}, {'label': 'Option 2', 'value': '2'}]
		
        Keyword Args: 
		multiple (bool): Whether the user can select multiple options. Defaults to False.
		min (number): The minimal amount of options that should be selected. Defaults to None.
		max (number): The maximum amount of options that should be selected. Defaults to None.
		initial_value ([]): The initial values of the selection. Defaults to [].
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The values/value selected by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_multiple_choice(label, options, **kwargs).run(button_text))


def read_nps(label: str, **kwargs):
  
    """Captures Net Promoter Score feedback with a 0-10 rating scale.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		min (int): Min value accepted by the input. Defaults to 0.
		max (int): Max value accepted by the input. Defaults to 10.
		min_hint (str): Text to display next to the min value. Defaults to "Not at all likely".
		max_hint (str): Text to display next to the max value. Defaults to "Extremely likely".
		initial_value (str): The initial value to display to the user. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_nps(label, **kwargs).run(button_text))


def read_number(label: str, **kwargs):
  
    """Collects numeric input with optional min/max limits.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (number): The initial value to display to the user. Defaults to 0.
		placeholder (str): The placeholder text to display to the user. Defaults to "".
		min (float): Min value accepted by the input. Defaults to None.
		max (float): Max value accepted by the input. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_number(label, **kwargs).run(button_text))


def read_number_slider(label: str, **kwargs):
  
    """Provides a slider for selecting numerical values within a range.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (float): The initial value to display to the user. Defaults to 0.
		min (float): Min value accepted by the input. Defaults to None.
		max (float): Max value accepted by the input. Defaults to None.
		step (float): The value to be incremented or decremented while using the input button. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_number_slider(label, **kwargs).run(button_text))


def read_pandas_row_selection(df: "DataFrame", **kwargs):
  
    """Enables selection of rows from a displayed pandas DataFrame table.

        Position Args: 
		df ("DataFrame"): The pandas dataframe to be displayed
		
        Keyword Args: 
		display_index (bool): Whether to show a index column. Defaults to False.
		label (str): The label to display to the user
		initial_value (list): The initial value of the selection. Defaults to []
		multiple (bool): Whether the user will be allowed to select multiple rows. Defaults to False.
		min (number): The minimal amount of options that should be selected. Defaults to None.
		max (number): The maximum amount of options that should be selected. Defaults to None.
		page_size (number): The number of rows to display per page. Defaults to 10.
		full_width (bool): Whether the input should use full screen width. Defaults to True.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The list of selected rows
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_pandas_row_selection(df, **kwargs).run(button_text))


def read_password(label: str, **kwargs):
  
    """A password field, including strength validation options.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		placeholder (str): The placeholder text to display to the user. Defaults to "".
		lowercase_required (bool): Whether the input must have at least one lowercase character. Defaults to True.
		uppercase_required (bool): Whether the input must have at least one uppercase character. Defaults to True.
		special_required (bool): Whether the input must have at least one special character. Defaults to True.
		digit_required (bool): Whether the input must have at least one digit. Defaults to True.
		min_length (int): Minimum length of the password. Defaults to 8.
		max_length (int): Maximum length of the password. Defaults to None.
		size (int): Size of the password. Defaults to None.
		pattern (str): A regex pattern for the accepted password. Defaults to None.
		autocomplete (str): The autocomplete HTML attribute. Defaults to "current-password".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_password(label, **kwargs).run(button_text))


def read_phone(label: str, **kwargs):
  
    """A phone inputs with country code and national number.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (dict): The initial value to display to the user. It contains two keys: 'country_code' (string with optional + sign or number) and 'national_number' (str). Ex: {'country_code': '55', 'national_number': '21999990000'}.
		placeholder (str): The placeholder text to display in the national number input. Defaults to "".
		invalid_message (str): The message to display when the input is invalid
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A dict containing the value entered by the user ({"country_code": str, "national_number": str})
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_phone(label, **kwargs).run(button_text))


def read_rating(label: str, **kwargs):
  
    """Collects user feedback with emoji icons.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (int): The initial value to display to the user. Defaults to 0.
		max (float): Max value accepted by the input. Defaults to None.
		char (str): Which char should be displayed as icon?
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_rating(label, **kwargs).run(button_text))


def read_richtext(label: str, **kwargs):
  
    """Offers a rich text editor for formatted textual input.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		placeholder (str): The placeholder text to display to the user. Defaults to "Your rich text here".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The rich text value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_richtext(label, **kwargs).run(button_text))


def read_tag(label: str, **kwargs):
  
    """Enables input of tags or keywords with autocomplete suggestions.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (list): The initial value to display to the user. Defaults to [].
		placeholder (str): The placeholder text to display to the user. Defaults to "".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The values entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_tag(label, **kwargs).run(button_text))


def read(label: str, **kwargs):
  
    """Collects plain text input with customizable placeholders and validation.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		placeholder (str): The placeholder text to display to the user. Defaults to "Placeholder".
		mask (str): A mask to apply to the input. Defaults to None.
		max_length (int): The maximum length of the input. Defaults to None.
		min_length (int): The minimum length of the input. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read(label, **kwargs).run(button_text))


def read_textarea(label: str, **kwargs):
  
    """Provides a text area for multi-line text input.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		placeholder (str): The placeholder text to display to the user. Defaults to "".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_textarea(label, **kwargs).run(button_text))


def read_time(label: str, **kwargs):
  
    """Allows selection of a specific time.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str or datetime.time): The initial value to display to the user. Defaults to "00:00".
		format (str): Whether the input is in the format 24hs or AM/PM. Defaults to "24hs".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A datetime.time object representing the value entered by the user
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_time(label, **kwargs).run(button_text))


def read_toggle(label: str, **kwargs):
  
    """Offers a toggle switch between two defined options.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		on_text (str): Text of On Toggle option
		off_text (str): Text of Off Toggle option
		initial_value (bool): Initial value of the toggle
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The toggle value
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_toggle(label, **kwargs).run(button_text))


def read_video(label: str, **kwargs):
  
    """Facilitates video file uploads with preview and size validation.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (Union[str, io.IOBase]): The initial value to display to the user. Defaults to "".
		multiple (bool): Whether the user will be allowed to upload multiple files. Defaults to False.
		max_file_size (float): Maximum size allowed to be transfered in total in MB.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A dict containing the video uploaded by the user: FileResponse(path: Path, file: BufferedReader). If the multiple flag is set as True, it might contain a list of FileResponses.
    """

    
  
    button_text = kwargs.get("button_text", "i18n_next_action")
    return get_single_value(Page().read_video(label, **kwargs).run(button_text))

